/*
 * iAirZoning
 * Copyright 2020 José Antonio Jiménez Campos
 */

#include <stdio.h>
#include <esp/uart.h>
#include <esp8266.h>
#include <FreeRTOS.h>
#include <espressif/esp_common.h>
#include <task.h>

#include <homekit/characteristics.h>
#include "header.h"
#include "types.h"

#include "gate_control.h"

/*
 Función que selecciona el canal de la interfaz MCP23017 en base al
 GPIO dado. De 0 a 7 canal 0; de 8 a 15, canal 1.
 @param gpio        GPIO de la interfaz MCP23017
 @param value       Valor a establecer en el GPIO (0: LOW, 1: HIGH)
 */
static void mcp_gpio_write_alt(const uint8_t gpio, const bool value) {
    if (gpio < 8) {
        mcp_gpio_write(MCP_CHANNEL_A, gpio, value);
    } else {
        mcp_gpio_write(MCP_CHANNEL_B, gpio - 8, value);
    }
}

/*
 Tarea encargada de abrir o cerrar una compuerta motorizada
 @param args        Termostato HomeKit
 */
static void gate_open_close_task(void* args) {
    homekit_thermostat_t *homekit_thermostat = args;

    if (homekit_thermostat->gate_status == THERMOSTAT_GATE_MUST_BE_CLOSED) {
        homekit_thermostat->gate_status = THERMOSTAT_GATE_CLOSING;
        
        INFO("[Room %i] Closing gate for %i ms",
             homekit_thermostat->room,
             homekit_thermostat->time_open_close_ms);
        mcp_gpio_write_alt(homekit_thermostat->open_mcp_gpio, 1);
        mcp_gpio_write_alt(homekit_thermostat->close_mcp_gpio, 0);
        vTaskDelay(MS_TO_TICK(homekit_thermostat->time_open_close_ms));
        mcp_gpio_write_alt(homekit_thermostat->close_mcp_gpio, 1);
        
        homekit_thermostat->gate_status = THERMOSTAT_GATE_CLOSED;
        INFO("[Room %i] Gate closed",
             homekit_thermostat->room);
        
    } else if (homekit_thermostat->gate_status == THERMOSTAT_GATE_MUST_BE_OPENED) {
        homekit_thermostat->gate_status = THERMOSTAT_GATE_OPENING;
        
        INFO("[Room %i] Opening gate for %i ms",
             homekit_thermostat->room,
             homekit_thermostat->time_open_close_ms);
        mcp_gpio_write_alt(homekit_thermostat->close_mcp_gpio, 1);
        mcp_gpio_write_alt(homekit_thermostat->open_mcp_gpio, 0);
        vTaskDelay(MS_TO_TICK(homekit_thermostat->time_open_close_ms));
        mcp_gpio_write_alt(homekit_thermostat->open_mcp_gpio, 1);
        
        homekit_thermostat->gate_status = THERMOSTAT_GATE_OPENED;
        INFO("[Room %i] Gate opened",
             homekit_thermostat->room);
    }
    
    homekit_thermostat->gate_task_handle = NULL;
    vTaskDelete(NULL);
}

/*
 Función que recibe la orden de abrir o cerrar una compuerta, y decide si hay que
 activar o no los actuadores correspondientes en base al estado actual
 @param homekit_thermostat      Termostato sobre el cual se va a abrir o cerrar la compuerta
 @param target_status           Estado final que debe tener la compuerta (false: cerrada, true: abierta)
 */
void gate_open_close(homekit_thermostat_t* homekit_thermostat, const bool target_status) {
    INFO("[Room %i] Gate current status = %i",
         homekit_thermostat->room,
         homekit_thermostat->gate_status);
    INFO("[Room %i] Gate target status  = %i",
         homekit_thermostat->room,
         target_status);
    
    void run_gate_task() {
        if (homekit_thermostat->gate_task_handle) {
            INFO("[Room %i] Removing actual gate task",
                 homekit_thermostat->room);
            vTaskDelete(homekit_thermostat->gate_task_handle);
        }
        
        xTaskCreate(gate_open_close_task,
                    "gate_open_close_task",
                    READ_SENSORS_TASK_SIZE,
                    homekit_thermostat,
                    OPEN_CLOSE_GATE_TASK_PRIORITY,
                    &homekit_thermostat->gate_task_handle);
    }
    
    if (target_status == THERMOSTAT_GATE_CLOSED &&
        homekit_thermostat->gate_status != THERMOSTAT_GATE_CLOSED &&
        homekit_thermostat->gate_status != THERMOSTAT_GATE_CLOSING) {
        
        homekit_thermostat->gate_status = THERMOSTAT_GATE_MUST_BE_CLOSED;
        run_gate_task();
        
    } else if (target_status == THERMOSTAT_GATE_OPENED &&
               homekit_thermostat->gate_status != THERMOSTAT_GATE_OPENED &&
               homekit_thermostat->gate_status != THERMOSTAT_GATE_OPENING) {
        
        homekit_thermostat->gate_status = THERMOSTAT_GATE_MUST_BE_OPENED;
        run_gate_task();
    }
}

