/*
* iAirZoning
* Copyright 2020 José Antonio Jiménez Campos
*/

#include <stdio.h>
#include <esp/uart.h>
#include <esp8266.h>
#include <FreeRTOS.h>
#include <espressif/esp_common.h>
#include <task.h>

#include "header.h"

#include "irda.h"

static uint8_t ir_tx_gpio = 255;
static uint8_t ir_tx_freq;
static uint16_t ir_tx_packet_len = 0;
static bool ir_true;
static bool ir_false;
static bool ir_tx_is_running = false;

static uint16_t *ir_code;
static uint16_t ir_protocol[7];

typedef struct _ir_task_args {
    uint8_t* command;
    uint16_t command_len;
} ir_task_args_t;

/*
 Tarea encargada de transmitir el código IrDA
 @param args      Argumento de la tarea, que realmente es del tipo ir_task_args_t
 */
static void ir_tx_task(void* args) {
    INFO("[IrDA] Sending signal");
    
    ir_task_args_t* ir_task_args = args;

    ir_code[0] = ir_protocol[0];
    ir_code[1] = ir_protocol[1];
    ir_code[ir_tx_packet_len - 1] = ir_protocol[6];
    
    // Decoding BIT0 BIT1 command
    uint16_t ir_code_index = 2;
    for (uint16_t i = 0; i < ir_task_args->command_len; i++) {
        for (uint16_t j = 0; j < ir_task_args->command[i]; j++) {
            if (i % 2 == 0) {
                ir_code[ir_code_index] = ir_protocol[2];
                ir_code_index++;
                ir_code[ir_code_index] = ir_protocol[3];
                ir_code_index++;
            } else {
                ir_code[ir_code_index] = ir_protocol[4];
                ir_code_index++;
                ir_code[ir_code_index] = ir_protocol[5];
                ir_code_index++;
            }
        }
    }
    
    free(ir_task_args);

#ifdef IAIRZONING_DEBUG
    for (uint16_t i = 0; i < ir_tx_packet_len; i++) {
        printf("%s%5d ", i & 1 ? "-" : "+", ir_code[i]);
        if (i % 16 == 15) {
            printf("\n");
        }
    }
    printf("\n");
#endif // IAIRZONING_DEBUG

    // IR TRANSMITTER
    uint32_t start;
    
    do {
        vTaskDelay(MS_TO_TICK(80));
    } while (ir_tx_is_running);
    
    ir_tx_is_running = true;

    for (uint8_t repeats = 0; repeats < 10; repeats++){
        for (uint16_t i = 0; i < ir_tx_packet_len; i++) {
            if (ir_code[i] > 0) {
                if (i & 1) {    // Space
                    gpio_write(ir_tx_gpio, ir_false);
                    sdk_os_delay_us(ir_code[i]);
                } else {        // Mark
                    start = sdk_system_get_time();
                    while ((sdk_system_get_time() - start) < ir_code[i]) {
                        gpio_write(ir_tx_gpio, ir_true);
                        sdk_os_delay_us(ir_tx_freq);
                        gpio_write(ir_tx_gpio, ir_false);
                        sdk_os_delay_us(ir_tx_freq);
                    }
                }
            }
        }
        
        vTaskDelay(MS_TO_TICK(100));
    }
    
    gpio_write(ir_tx_gpio, ir_false);
 
    ir_tx_is_running = false;

    vTaskDelete(NULL);
}

/*
 Prepara el comando IrDA para ser enviado posteriormente por la tarea ir_tx_task.
 @param command         Puntero del array de uint8_t que codifica el comando a enviar
 @param command_len     Tamaño del array command
 */
void ir_tx_send(uint8_t* command, const uint16_t command_len) {
    if (ir_tx_gpio != 255 && ir_tx_packet_len != 0) {
        ir_task_args_t *ir_task_args = malloc(sizeof(ir_task_args_t));
        ir_task_args->command = command;
        ir_task_args->command_len = command_len;
        
        xTaskCreate(ir_tx_task, "ir_tx_task", IR_TX_TASK_SIZE, ir_task_args, IR_TX_TASK_PRIORITY, NULL);
    }
}

/*
 Prepara el GPIO encargado de manejar el LED Ir
 @param gpio        Número de GPIO donde está conectado el LED Ir
 @param inverted    Indica si el comportamiento del LED Ir es inverso al estado del GPIO
 @param freq        Frecuencia en KHz de las marcas de la yrama Ir
 */
void ir_tx_init(const uint8_t gpio, const bool inverted, const uint8_t freq) {
    ir_tx_gpio = gpio;
    ir_tx_freq = 1000 / freq / 2;
    
    ir_true = true ^ inverted;
    ir_false = false ^ inverted;
    
    gpio_enable(ir_tx_gpio, GPIO_OUTPUT);
    gpio_write(ir_tx_gpio, ir_false);
}

/*
 Establece el protocolo IrDA a usar
 @param header_mark     Tiempo de duración en microsegundos de la marca de la cabecera
 @param header_space    Tiempo de duración en microsegundos del espacio de la cabecera
 @param bit0_mark       Tiempo de duración en microsegundos de la marca del bit 0
 @param bit0_space      Tiempo de duración en microsegundos del espacio del bit 0
 @param bit1_mark       Tiempo de duración en microsegundos de la marca del bit 1
 @param bit1_space      Tiempo de duración en microsegundos del espacio del bit 1
 @param footer_mark     Tiempo de duración en microsegundos de la marca del final de la trama Ir
 @param packet_len      Tamaño de la trama Ir
 */
void ir_tx_set_protocol(const uint16_t header_mark, const uint16_t header_space,
                        const uint16_t bit0_mark, const uint16_t bit0_space,
                        const uint16_t bit1_mark, const uint16_t bit1_space,
                        const uint16_t footer_mark,
                        const uint16_t packet_len) {
    ir_protocol[0] = header_mark;
    ir_protocol[1] = header_space;
    ir_protocol[2] = bit0_mark;
    ir_protocol[3] = bit0_space;
    ir_protocol[4] = bit1_mark;
    ir_protocol[5] = bit1_space;
    ir_protocol[6] = footer_mark;
    
    ir_tx_packet_len = packet_len;
    ir_code = malloc(sizeof(uint16_t) * ir_tx_packet_len);
}
