/*
 * iAirZoning
 * Copyright 2020 José Antonio Jiménez Campos
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <esp/uart.h>
#include <esp8266.h>
#include <FreeRTOS.h>
#include <espressif/esp_wifi.h>
#include <espressif/esp_common.h>
#include <task.h>

#include <etstimer.h>
#include <esplibs/libmain.h>

#include <homekit/homekit.h>
#include <homekit/characteristics.h>

#include <dht.h>

#include "header.h"
#include "types.h"

#include "user_config.h"

#include "gate_control.h"
#include "irda.h"

/*
 Función obligatoria de identificación de HomeKit. No tiene ninguna utilidad real
 para este proyecto.
 @param _value      Valor HomeKit, requerido por la API pero sin utilidad real
 */
void identify(homekit_value_t _value) {
    INFO("[HomeKit] Identify");
}

uint8_t wifi_status                         = WIFI_STATUS_CONNECTED_STAGE_2;
uint8_t wifi_channel                        = 0;

uint8_t ac_main_mode                        = 255;
uint8_t current_ac_main_ir_mode             = 255;

ETSTimer wifi_watchdog_timer;
ETSTimer set_zone_settings_timer;
ETSTimer read_sensors_timer;

/*
 Características HomeKit de información de cada accesorio a crear
 */
homekit_characteristic_t name               = HOMEKIT_CHARACTERISTIC_(NAME, "iAirZoning");
homekit_characteristic_t serial             = HOMEKIT_CHARACTERISTIC_(SERIAL_NUMBER, NULL);
homekit_characteristic_t manufacturer       = HOMEKIT_CHARACTERISTIC_(MANUFACTURER, "José A. Jiménez Campos");
homekit_characteristic_t model              = HOMEKIT_CHARACTERISTIC_(MODEL, "UOC TFG 2020");
homekit_characteristic_t identify_function  = HOMEKIT_CHARACTERISTIC_(IDENTIFY, identify);
homekit_characteristic_t firmware           = HOMEKIT_CHARACTERISTIC_(FIRMWARE_REVISION, FIRMWARE_VERSION);
homekit_characteristic_t hap_version        = HOMEKIT_CHARACTERISTIC_(VERSION, "1.1.0");

homekit_server_config_t config;

homekit_thermostat_t *homekit_thermostats   = NULL;

char serial_value[13];

/*
 Inicia la conexión WiFi
 */
void wifi_connect() {
    INFO("[WiFi] Connecting...");
    
    char *wifi_ssid = WIFI_USER_CONFIG_SSID;
    char *wifi_password = WIFI_USER_CONFIG_PASSWD;

    struct sdk_station_config sta_config;
               
    memset(&sta_config, 0, sizeof(sta_config));
    strncpy((char *)sta_config.ssid, wifi_ssid, sizeof(sta_config.ssid));
    
    sta_config.ssid[sizeof(sta_config.ssid) - 1] = 0;

    if (wifi_password) {
       strncpy((char *)sta_config.password, wifi_password, sizeof(sta_config.password));
    }

    sdk_wifi_station_set_config(&sta_config);
    sdk_wifi_station_connect();
    sdk_wifi_station_set_auto_connect(true);
}

/*
 Desconecta la WiFi y borra la configuración actual, para prevenir usos de caché residual.
 Para ello establece una configuración falsa
 */
void wifi_reset() {
    INFO("[WiFi] Resetting...");
    
    sdk_wifi_station_disconnect();
    
    struct sdk_station_config sta_config;
    memset(&sta_config, 0, sizeof(sta_config));

    strncpy((char *)sta_config.ssid, "none", sizeof(sta_config.ssid));
    sta_config.ssid[sizeof(sta_config.ssid) - 1] = 0;
    
    sdk_wifi_station_set_config(&sta_config);
    sdk_wifi_station_connect();
    sdk_wifi_station_set_auto_connect(true);
    
    INFO("[WiFi] Resetted");
}

#ifdef IAIRZONING_DEBUG
/*
 Monitoriza la memoria HEAP libre y el tamaño máximo de bloque
 */
void memory_check() {
    int32_t size = xPortGetFreeHeapSize();
    INFO("*** Free heap = %i", size);
    
    char* space = NULL;
    
    while (!space && size > 0) {
        space = malloc(size);
        size -= 4;
    }
    
    free(space);
    
    INFO("*** Max chunk = %i", size + 4);
}
#endif // IAIRZONING_DEBUG

/*
 Monitoriza la conexión WiFi para reiniciarla en caso de desconexión, y reenvía los anuncios
 mDNS de HomeKit una vez que se ha vuelto a conectar. También lo hace si la WiFi salta a otro
 punto de acceso con distinto canal (Por ejemplo, en redes de mallado).
 */
void wifi_watchdog() {
    if (wifi_status == WIFI_STATUS_CONNECTED_STAGE_2) {
        uint8_t current_channel = sdk_wifi_get_channel();
        if (wifi_channel != current_channel) {
            wifi_status = WIFI_STATUS_CONNECTED_STAGE_1;
            ERROR("[WiFi] New channel: %i",
                  current_channel);
            wifi_channel = current_channel;
            homekit_mdns_announce();
        }
    } else if (wifi_status == WIFI_STATUS_DISCONNECTED) {
        wifi_status = WIFI_STATUS_CONNECTING;
        INFO("[WiFi] Connecting...");
        wifi_connect();
        
    } else if (sdk_wifi_station_get_connect_status() == STATION_GOT_IP) {
        if (wifi_status == WIFI_STATUS_CONNECTING) {
            wifi_status = WIFI_STATUS_CONNECTED_STAGE_1;
            INFO("[WiFi] Connected");
            homekit_mdns_announce();
            
        } else if (wifi_status == WIFI_STATUS_CONNECTED_STAGE_1) {
            wifi_status = WIFI_STATUS_CONNECTED_STAGE_2;
            wifi_channel = sdk_wifi_get_channel();
            INFO("[WiFi] mDNS reannounced");
            homekit_mdns_announce();
        }
        
    } else {
        wifi_status = WIFI_STATUS_DISCONNECTED;
        ERROR("[WiFi] Disconnected");
        wifi_reset();
    }
    
#ifdef IAIRZONING_DEBUG
    memory_check();
#endif // IAIRZONING_DEBUG
    
}

/*
 Devuelve el termostato en base a una de sus características
 @param ch      Característica HomeKit a buscar
 @return        Termostato. NULL si no existe
 */
homekit_thermostat_t* homekit_thermostat_find_by_ch(homekit_characteristic_t* ch) {
    homekit_thermostat_t *homekit_thermostat = homekit_thermostats;
    
    while (homekit_thermostat                       &&
           homekit_thermostat->active               != ch &&
           homekit_thermostat->current_temp         != ch &&
           homekit_thermostat->current_state        != ch &&
           homekit_thermostat->target_state         != ch &&
           homekit_thermostat->target_heating_temp  != ch &&
           homekit_thermostat->target_cooling_temp  != ch &&
           homekit_thermostat->current_hum          != ch) {
        homekit_thermostat = homekit_thermostat->next;
    }

    return homekit_thermostat;
}

/*
 Devuelve el termostato en base al número de habitación
 @param room    Número de habitación a buscar
 @return        Termostato. NULL si no existe
 */
homekit_thermostat_t* homekit_thermostat_find_by_room(uint8_t room) {
    homekit_thermostat_t *homekit_thermostat = homekit_thermostats;
    
    while (homekit_thermostat &&
           homekit_thermostat->room != room) {
        homekit_thermostat = homekit_thermostat->next;
    }

    return homekit_thermostat;
}

/*
 Notifica en grupo al servidor HomeKit de todas las características de un termostato
 @param homekit_thermostat      Termostato del que hay que notificar sus características
 */
void homekit_thermostat_notify(homekit_thermostat_t* homekit_thermostat) {
    INFO("[HomeKit] Send changes over HAP");
    homekit_characteristic_notify(homekit_thermostat->active,
                                  homekit_thermostat->active->value);
    
    homekit_characteristic_notify(homekit_thermostat->current_temp,
                                  homekit_thermostat->current_temp->value);
    
    homekit_characteristic_notify(homekit_thermostat->display_units,
                                  homekit_thermostat->display_units->value);
    
    homekit_characteristic_notify(homekit_thermostat->current_state,
                                  homekit_thermostat->current_state->value);
    
    homekit_characteristic_notify(homekit_thermostat->target_state,
                                  homekit_thermostat->target_state->value);
    
    homekit_characteristic_notify(homekit_thermostat->target_heating_temp,
                                  homekit_thermostat->target_heating_temp->value);
    
    homekit_characteristic_notify(homekit_thermostat->target_cooling_temp,
                                  homekit_thermostat->target_cooling_temp->value);
    
    homekit_characteristic_notify(homekit_thermostat->current_hum,
                                  homekit_thermostat->current_hum->value);
}

/*
 Establece el valor de una característica en HomeKit y notifica
 @param ch      Característica HomeKit a establecer
 @param value   Nuevo valor que ha de tomar la característica
 */
void hkc_setter(homekit_characteristic_t* ch, const homekit_value_t value) {
    INFO("[HomeKit] Default setter");
    ch->value = value;
    homekit_thermostat_notify(homekit_thermostat_find_by_ch(ch));
}

/*
 Función que establece el estado de las compuertas y de la máquina de A/C en base a
 la información de cada termostato
 */
void set_zone_settings() {
    INFO("[iAirZoning] Calculating zone settings");

    uint8_t final_ac_main_ir_mode = current_ac_main_ir_mode;
    
    // Fix impossible cases
    homekit_thermostat_t* homekit_thermostat = homekit_thermostats;
    while (homekit_thermostat) {
        switch (homekit_thermostat->mode) {
            case THERMOSTAT_HEATER_ON:
            case THERMOSTAT_HEATER_IDLE:
            case THERMOSTAT_HEATER_FORCE_IDLE:
                if (ac_main_mode == AC_MAIN_MODE_COOLER) {
                    homekit_thermostat->mode = THERMOSTAT_OFF;
                    hkc_setter(homekit_thermostat->active, HOMEKIT_UINT8(THERMOSTAT_MODE_OFF));
                } else {
                    ac_main_mode = AC_MAIN_MODE_HEATER;
                }
                break;
                
            case THERMOSTAT_COOLER_ON:
            case THERMOSTAT_COOLER_IDLE:
            case THERMOSTAT_COOLER_FORCE_IDLE:
                if (ac_main_mode == AC_MAIN_MODE_HEATER) {
                    homekit_thermostat->mode = THERMOSTAT_OFF;
                    hkc_setter(homekit_thermostat->active, HOMEKIT_UINT8(THERMOSTAT_MODE_OFF));
                } else {
                    ac_main_mode = AC_MAIN_MODE_COOLER;
                }
                break;
                
            default:    // THERMOSTAT_OFF
                // Do nothing
                break;
        }

        homekit_thermostat = homekit_thermostat->next;
    }
    
    bool thermostat_all_off = true;
    bool thermostat_all_idle = true;
    bool thermostat_all_soft_on = true;
    bool thermostat_force_idle = false;
    homekit_thermostat = homekit_thermostats;
    while (homekit_thermostat && thermostat_all_idle) {
        INFO("[iAirZoning] Thermostat Mode: %i", homekit_thermostat->mode);
        if (homekit_thermostat->mode != THERMOSTAT_OFF) {
            thermostat_all_off = false;
            
            if (homekit_thermostat->mode == THERMOSTAT_HEATER_ON ||
                homekit_thermostat->mode == THERMOSTAT_COOLER_ON) {
                thermostat_all_idle = false;
                thermostat_all_soft_on = false;
                
            } else if (homekit_thermostat->mode == THERMOSTAT_HEATER_SOFT_ON ||
                       homekit_thermostat->mode == THERMOSTAT_COOLER_SOFT_ON) {
                thermostat_all_idle = false;

            } else if (homekit_thermostat->mode == THERMOSTAT_HEATER_FORCE_IDLE ||
                       homekit_thermostat->mode == THERMOSTAT_COOLER_FORCE_IDLE) {
                thermostat_force_idle = true;
                
            }
        }
        
        homekit_thermostat = homekit_thermostat->next;
    }
    
    INFO("[iAirZoning] All OFF: %i, all IDLE: %i, all soft ON: %i, force IDLE: %i",
         thermostat_all_off,
         thermostat_all_idle,
         thermostat_all_soft_on,
         thermostat_force_idle);
    
    if (thermostat_all_off) {
        if (ac_main_mode != AC_MAIN_MODE_OFF) {
            INFO("[iAirZoning] Total OFF");
            ac_main_mode = AC_MAIN_MODE_OFF;
            final_ac_main_ir_mode = AC_MAIN_IR_OFF;
            
            // Open all gates
            homekit_thermostat = homekit_thermostats;
            while (homekit_thermostat) {
                gate_open_close(homekit_thermostat, THERMOSTAT_GATE_OPENED);
                
                homekit_thermostat = homekit_thermostat->next;
            }
        }
        
    } else {
        if (thermostat_all_idle) {
            if (ac_main_mode == AC_MAIN_MODE_HEATER) {
                if (thermostat_force_idle) {
                    final_ac_main_ir_mode = AC_MAIN_IR_HEATER_FORCE_IDLE;
                } else {
                    final_ac_main_ir_mode = AC_MAIN_IR_HEATER_IDLE;
                }
            } else {
                if (thermostat_force_idle) {
                    final_ac_main_ir_mode = AC_MAIN_IR_COOLER_FORCE_IDLE;
                } else {
                    final_ac_main_ir_mode = AC_MAIN_IR_COOLER_IDLE;
                }
            }
        } else if (thermostat_all_soft_on) {
            if (ac_main_mode == AC_MAIN_MODE_HEATER) {
                final_ac_main_ir_mode = AC_MAIN_IR_HEATER_SOFT_ON;
            } else {
                final_ac_main_ir_mode = AC_MAIN_IR_COOLER_SOFT_ON;
            }
        } else {
            if (ac_main_mode == AC_MAIN_MODE_HEATER) {
                final_ac_main_ir_mode = AC_MAIN_IR_HEATER_ON;
            } else {
                final_ac_main_ir_mode = AC_MAIN_IR_COOLER_ON;
            }
        }
        
        homekit_thermostat = homekit_thermostats;
        while (homekit_thermostat) {
            switch (homekit_thermostat->mode) {
                case THERMOSTAT_HEATER_ON:
                case THERMOSTAT_COOLER_ON:
                case THERMOSTAT_HEATER_SOFT_ON:
                case THERMOSTAT_COOLER_SOFT_ON:
                    gate_open_close(homekit_thermostat, THERMOSTAT_GATE_OPENED);
                    break;
                    
                case THERMOSTAT_HEATER_IDLE:
                case THERMOSTAT_COOLER_IDLE:
                case THERMOSTAT_HEATER_FORCE_IDLE:
                case THERMOSTAT_COOLER_FORCE_IDLE:
                    if (thermostat_all_idle) {
                        gate_open_close(homekit_thermostat, THERMOSTAT_GATE_OPENED);
                    } else {
                        gate_open_close(homekit_thermostat, THERMOSTAT_GATE_CLOSED);
                    }
                    break;
                    
                default:    // THERMOSTAT_OFF
                    gate_open_close(homekit_thermostat, THERMOSTAT_GATE_CLOSED);
                    break;
            }

            homekit_thermostat = homekit_thermostat->next;
        }
    }
    
    if (final_ac_main_ir_mode != current_ac_main_ir_mode) {
        INFO("[iAirZoning] Sending IR to A/C main (%i)", final_ac_main_ir_mode);
        current_ac_main_ir_mode = final_ac_main_ir_mode;

        uint8_t* command = NULL;
        uint16_t len;

        switch (current_ac_main_ir_mode) {
            case AC_MAIN_IR_HEATER_ON:
                command = ac_main_ir_heater_on;
                len = sizeof(ac_main_ir_heater_on);
                break;
            
            case AC_MAIN_IR_HEATER_SOFT_ON:
            case AC_MAIN_IR_HEATER_IDLE:
                command = ac_main_ir_heater_idle;
                len = sizeof(ac_main_ir_heater_idle);
                break;
                
            case AC_MAIN_IR_HEATER_FORCE_IDLE:
                command = ac_main_ir_heater_force_idle;
                len = sizeof(ac_main_ir_heater_force_idle);
                break;
                
            case AC_MAIN_IR_COOLER_ON:
                command = ac_main_ir_cooler_on;
                len = sizeof(ac_main_ir_cooler_on);
                break;
            
            case AC_MAIN_IR_COOLER_SOFT_ON:
            case AC_MAIN_IR_COOLER_IDLE:
                command = ac_main_ir_cooler_idle;
                len = sizeof(ac_main_ir_cooler_idle);
                break;
                
            case AC_MAIN_IR_COOLER_FORCE_IDLE:
                command = ac_main_ir_cooler_force_idle;
                len = sizeof(ac_main_ir_cooler_force_idle);
                break;
                
            default:    // AC_MAIN_IR_OFF
                command = ac_main_ir_off;
                len = sizeof(ac_main_ir_off);
                break;
        }

        ir_tx_send(command, len);
    }
}

/*
 Función que retrasa la ejecución de set_zone_settings para evitar ejecuciones
 innecesarias.
 */
inline void set_zone_settings_delayed() {
    sdk_os_timer_arm(&set_zone_settings_timer, SET_ZONE_SETTINGS_DELAY_MS, false);
}

/*
 Función que actualiza la característica HomeKit correspondiende y lanza un timer
 que ehecutará la función update_th_delayed correspondiente al termostato. Esto se
 hace para evitar iteraciones innecesarias cuando HomeKit actualiza varias características
 al mismo tiempo.
 @param ch      Característica HomeKit a establecer
 @param value   Nuevo valor que ha de tomar la característica
*/
void update_th(homekit_characteristic_t* ch, const homekit_value_t value) {
    INFO("[HomeKit] Thermostat set value");
    
    ch->value = value;
    
    sdk_os_timer_arm(&homekit_thermostat_find_by_ch(ch)->update_th_timer, UPDATE_TH_DELAY_MS, false);
}

/*
 Función de actualización del termostato y lógica de su estado en función de la temperatura a alcanzar
 @param args    Thermostato
 */
void update_th_delayed(void* args) {
    INFO("[iAirZoning] Thermostat working...");
    homekit_thermostat_t *homekit_thermostat = args;

    void heating(const float deadband, const float deadband_soft_on, const float deadband_force_idle) {
        if (homekit_thermostat->current_temp->value.float_value < (homekit_thermostat->target_heating_temp->value.float_value - deadband - deadband_soft_on)) {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_HEATER;
            homekit_thermostat->mode = THERMOSTAT_HEATER_ON;
            
        } else if (homekit_thermostat->current_temp->value.float_value < (homekit_thermostat->target_heating_temp->value.float_value - deadband)) {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_HEATER;
            homekit_thermostat->mode = THERMOSTAT_HEATER_SOFT_ON;
            
        } else if (homekit_thermostat->current_temp->value.float_value < (homekit_thermostat->target_heating_temp->value.float_value + deadband)) {
            if (homekit_thermostat->current_state->value.int_value == THERMOSTAT_MODE_HEATER) {
                if (THERMOSTAT_DEADBAND_SOFT_ON > 0.000f) {
                    homekit_thermostat->mode = THERMOSTAT_HEATER_SOFT_ON;
                } else {
                    homekit_thermostat->mode = THERMOSTAT_HEATER_ON;
                }
                
            } else {
                homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_IDLE;
                homekit_thermostat->mode = THERMOSTAT_HEATER_IDLE;
            }
            
        } else if (homekit_thermostat->current_temp->value.float_value >= (homekit_thermostat->target_heating_temp->value.float_value + deadband_force_idle) &&
                   THERMOSTAT_DEADBAND_FORCE_IDLE > 0.000f) {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_IDLE;
            homekit_thermostat->mode = THERMOSTAT_HEATER_FORCE_IDLE;
            
        } else {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_IDLE;
            if (THERMOSTAT_DEADBAND_FORCE_IDLE == 0.000f ||
                homekit_thermostat->mode != THERMOSTAT_HEATER_FORCE_IDLE) {
                homekit_thermostat->mode = THERMOSTAT_HEATER_IDLE;
            }
        }
    }
    
    void cooling(const float deadband, const float deadband_soft_on, const float deadband_force_idle) {
        if (homekit_thermostat->current_temp->value.float_value > (homekit_thermostat->target_cooling_temp->value.float_value + deadband + deadband_soft_on)) {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_COOLER;
            homekit_thermostat->mode = THERMOSTAT_COOLER_ON;
            
        } else if (homekit_thermostat->current_temp->value.float_value > (homekit_thermostat->target_cooling_temp->value.float_value + deadband)) {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_COOLER;
            homekit_thermostat->mode = THERMOSTAT_COOLER_SOFT_ON;
            
        } else if (homekit_thermostat->current_temp->value.float_value > (homekit_thermostat->target_cooling_temp->value.float_value - deadband)) {
            if (homekit_thermostat->current_state->value.int_value == THERMOSTAT_MODE_COOLER) {
                if (THERMOSTAT_DEADBAND_SOFT_ON > 0.000f) {
                    homekit_thermostat->mode = THERMOSTAT_COOLER_SOFT_ON;
                } else {
                    homekit_thermostat->mode = THERMOSTAT_COOLER_ON;
                }
                
            } else {
                homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_IDLE;
                homekit_thermostat->mode = THERMOSTAT_COOLER_IDLE;
            }
            
        } else if (homekit_thermostat->current_temp->value.float_value <= (homekit_thermostat->target_cooling_temp->value.float_value - deadband_force_idle) &&
                   THERMOSTAT_DEADBAND_FORCE_IDLE > 0.000f) {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_IDLE;
            homekit_thermostat->mode = THERMOSTAT_COOLER_FORCE_IDLE;
            
        } else {
            homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_IDLE;
            if (THERMOSTAT_DEADBAND_FORCE_IDLE == 0.000f ||
                homekit_thermostat->mode != THERMOSTAT_COOLER_FORCE_IDLE) {
                homekit_thermostat->mode = THERMOSTAT_COOLER_IDLE;
            }
        }
    }
    
    if (homekit_thermostat->active->value.int_value) {
        if (homekit_thermostat->target_state->value.int_value == THERMOSTAT_TARGET_MODE_HEATER) {
            heating(THERMOSTAT_DEADBAND / 2.000f, THERMOSTAT_DEADBAND_SOFT_ON, THERMOSTAT_DEADBAND_FORCE_IDLE);
                    
        } else if (homekit_thermostat->target_state->value.int_value == THERMOSTAT_TARGET_MODE_COOLER) {
            cooling(THERMOSTAT_DEADBAND / 2.000f, THERMOSTAT_DEADBAND_SOFT_ON, THERMOSTAT_DEADBAND_FORCE_IDLE);
            
        } else {    // THERMOSTAT_TARGET_MODE_AUTO
            const float mid_target_temp = (homekit_thermostat->target_heating_temp->value.float_value + homekit_thermostat->target_cooling_temp->value.float_value) / 2.00f;
            
            bool is_heater = false;
            if (homekit_thermostat->current_state->value.int_value == THERMOSTAT_MODE_OFF) {
                if (homekit_thermostat->current_temp->value.float_value < mid_target_temp) {
                    is_heater = true;
                }
            } else if (homekit_thermostat->current_temp->value.float_value <= homekit_thermostat->target_heating_temp->value.float_value) {
                is_heater = true;
            } else if (homekit_thermostat->current_temp->value.float_value < homekit_thermostat->target_cooling_temp->value.float_value &&
                       (homekit_thermostat->mode == THERMOSTAT_HEATER_IDLE ||
                       homekit_thermostat->mode == THERMOSTAT_HEATER_ON ||
                       homekit_thermostat->mode == THERMOSTAT_HEATER_FORCE_IDLE ||
                       homekit_thermostat->mode == THERMOSTAT_HEATER_SOFT_ON)) {
                is_heater = true;
            }
            
            const float th_deadband_force_idle = homekit_thermostat->target_cooling_temp->value.float_value - mid_target_temp;
            const float th_deadband = th_deadband_force_idle / 1.500f;
            
            if (is_heater) {
                heating(th_deadband, th_deadband_force_idle - th_deadband, th_deadband_force_idle);
            } else {
                cooling(th_deadband, th_deadband_force_idle - th_deadband, th_deadband_force_idle);
            }
        }
        
    } else {
        homekit_thermostat->current_state->value.int_value = THERMOSTAT_MODE_OFF;
        homekit_thermostat->mode = THERMOSTAT_OFF;
    }

    homekit_thermostat_notify(homekit_thermostat);
    
    set_zone_settings_delayed();
}

/*
 Tarea encargada de leer los datos de los sensores y actualizar las
 características correspondientes en HomeKit (Temperatura y Humedad). Además,
 controla posibles errores en la lectura de los sensores.
 */
void read_sensors_task() {
    INFO("[Sensors] Reading data...");
    
    homekit_thermostat_t *homekit_thermostat = homekit_thermostats;
    
    while (homekit_thermostat) {
        float temperature_value, humidity_value;
        
        if (dht_read_float_data(DHT_TYPE_DHT22,
                                homekit_thermostat->sensor_gpio,
                                &humidity_value,
                                &temperature_value)) {
            homekit_thermostat->sensor_errors_count = 0;
            
            temperature_value += homekit_thermostat->temp_offset;
            humidity_value += homekit_thermostat->hum_offset;
            
            if (humidity_value < 0) {
                humidity_value = 0;
            } else if (humidity_value > 100) {
                humidity_value = 100;
            }
            
            INFO("[Room %i] Data from GPIO %i: Temp %.02f, Hum %.02f",
                 homekit_thermostat->room,
                 homekit_thermostat->sensor_gpio,
                 temperature_value, humidity_value);

        } else {
            homekit_thermostat->sensor_errors_count++;
            
            ERROR("[Room %i] Error reading sensor from GPIO %i (%i tries)",
                  homekit_thermostat->room,
                  homekit_thermostat->sensor_gpio,
                  homekit_thermostat->sensor_errors_count);

            if (homekit_thermostat->sensor_errors_count > TH_SENSOR_MAX_ALLOWED_ERRORS) {
                ERROR("[Room %i] Turning off thermostat...",
                      homekit_thermostat->room);
                
                temperature_value = 0;
                humidity_value = 0;
                homekit_thermostat->sensor_errors_count = 0;
                
                update_th(homekit_thermostat->active, HOMEKIT_UINT8(0));
                
            } else {
                // Keep last temp and hum reads
                temperature_value = homekit_thermostat->current_temp->value.float_value;
                humidity_value = homekit_thermostat->current_hum->value.float_value;
            }
        }
        
        if (homekit_thermostat->current_hum->value.float_value != humidity_value) {
            homekit_thermostat->current_hum->value = HOMEKIT_FLOAT(humidity_value);
        }
        
        if (homekit_thermostat->current_temp->value.float_value != temperature_value) {
            update_th(homekit_thermostat->current_temp, HOMEKIT_FLOAT(temperature_value));
        }

        homekit_thermostat = homekit_thermostat->next;
    }
    
    INFO("[Sensors] Reading data finished");
    
    vTaskDelete(NULL);
}

/*
 Función que se ejecuta cada TH_SENSOR_POLL_PERIOD_MS milisegundos y
 que crea la tarea encargada de leer los datos de los sensores.
 Si el accesorio está en proceso de emparejamiento, no se ejecuta
 para dejar libre toda la RAM posible.
 */
void read_sensors() {
    if (!homekit_is_pairing()) {
        xTaskCreate(read_sensors_task,
                    "read_sensors_task",
                    READ_SENSORS_TASK_SIZE,
                    NULL,
                    READ_SENSORS_TASK_PRIORITY,
                    NULL);
    }
}

/*
 Tarea de inicio
 */
void init_task() {
    INFO("Initial task...");
    
    INFO("[IrDA] Set GPIO %i", IR_LED_GPIO);
    ir_tx_init(IR_LED_GPIO,
               IR_LED_INVERTED,
               IR_LED_FREQ);
    
    INFO("[IrDA] Set Protocol");
    ir_tx_set_protocol(IR_HEADER_MARK, IR_HEADER_SPACE,
                       IR_BIT0_MARK, IR_BIT0_SPACE,
                       IR_BIT1_MARK, IR_BIT1_SPACE,
                       IR_FOOTER_MARK,
                       IR_PACKET_SIZE);
    
    INFO("Starting MCP 23017 interface. (errno = %i)",
         mcp_init(MCP_I2C_BUS,
                  MCP_SCL_GPIO,
                  MCP_SDA_GPIO,
                  INIT_MCP_A_OUTS,
                  INIT_MCP_B_OUTS));
    
    // Set MAC address to serial
    uint8_t macaddr[6];
    sdk_wifi_get_macaddr(STATION_IF, macaddr);
    snprintf(serial_value,
             13,
             "%02X%02X%02X%02X%02X%02X",
             macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4], macaddr[5]);
    serial.value = HOMEKIT_STRING(serial_value);
    INFO("[WiFi] MAC address: %s",
         serial_value);
    
    homekit_accessory_t **accessories = calloc(TOTAL_ROOMS + 2, sizeof(homekit_accessory_t*));
    
    /*
     Función interna para la creación del servicio HomeKit de información
     */
    void new_accessory(const uint8_t accessory, const uint8_t services) {
        accessories[accessory]                                  = calloc(1, sizeof(homekit_accessory_t));
        accessories[accessory]->id                              = accessory + 1;
        accessories[accessory]->services                        = calloc(services, sizeof(homekit_service_t*));
        
        accessories[accessory]->services[0]                     = calloc(1, sizeof(homekit_service_t));
        accessories[accessory]->services[0]->id                 = 1;
        accessories[accessory]->services[0]->type               = HOMEKIT_SERVICE_ACCESSORY_INFORMATION;
        accessories[accessory]->services[0]->characteristics    = calloc(7, sizeof(homekit_characteristic_t*));
        accessories[accessory]->services[0]->characteristics[0] = &name;
        accessories[accessory]->services[0]->characteristics[1] = &manufacturer;
        accessories[accessory]->services[0]->characteristics[2] = &serial;
        accessories[accessory]->services[0]->characteristics[3] = &model;
        accessories[accessory]->services[0]->characteristics[4] = &firmware;
        accessories[accessory]->services[0]->characteristics[5] = &identify_function;
        
        if (accessory == 0) {   // HAP Information
            INFO("[HomeKit] HAP v1.1.0");
            accessories[accessory]->services[1]                     = calloc(1, sizeof(homekit_service_t));
            accessories[accessory]->services[1]->id                 = 8;
            accessories[accessory]->services[1]->type               = HOMEKIT_SERVICE_HAP_INFORMATION;
            accessories[accessory]->services[1]->characteristics    = calloc(2, sizeof(homekit_characteristic_t*));
            accessories[accessory]->services[1]->characteristics[0] = &hap_version;
        }
    }

    INFO("\n--------------");
    INFO("\n[HomeKit] Creating Accessories");
    
    // Accessory Builder
    INFO("[HomeKit] Creating Bridge");
    new_accessory(0, 3);
    
    const float default_target_temp = (THERMOSTAT_SET_MIN_TEMP + THERMOSTAT_SET_MAX_TEMP) / 2;
    
    for(uint8_t i = TOTAL_ROOMS; i > 0; i--) {
        const uint8_t room_index = i - 1;
        
        INFO("\n[HomeKit] Creating room %i...",
             room_index);

        new_accessory(i, 4);

        // HomeKit Thermostat Characteristics
        homekit_characteristic_t *active                = NEW_HOMEKIT_CHARACTERISTIC(ACTIVE, 0, .setter_ex=update_th);
        homekit_characteristic_t *current_temp          = NEW_HOMEKIT_CHARACTERISTIC(CURRENT_TEMPERATURE, 0);
        homekit_characteristic_t *display_units         = NEW_HOMEKIT_CHARACTERISTIC(TEMPERATURE_DISPLAY_UNITS, 0, .setter_ex=hkc_setter);
        homekit_characteristic_t *current_state         = NEW_HOMEKIT_CHARACTERISTIC(CURRENT_HEATER_COOLER_STATE, 0);
        homekit_characteristic_t *target_state          = NEW_HOMEKIT_CHARACTERISTIC(TARGET_HEATER_COOLER_STATE, THERMOSTAT_TARGET_MODE_AUTO, .setter_ex=update_th);
        homekit_characteristic_t *target_heating_temp   = NEW_HOMEKIT_CHARACTERISTIC(HEATING_THRESHOLD_TEMPERATURE, default_target_temp - 1, .min_value=(float[]) {THERMOSTAT_SET_MIN_TEMP}, .max_value=(float[]) {THERMOSTAT_SET_MAX_TEMP}, .setter_ex=update_th);
        homekit_characteristic_t *target_cooling_temp   = NEW_HOMEKIT_CHARACTERISTIC(COOLING_THRESHOLD_TEMPERATURE, default_target_temp + 1, .min_value=(float[]) {THERMOSTAT_SET_MIN_TEMP}, .max_value=(float[]) {THERMOSTAT_SET_MAX_TEMP}, .setter_ex=update_th);
        
        accessories[i]->services[1]                     = calloc(1, sizeof(homekit_service_t));
        accessories[i]->services[1]->id                 = 8;
        accessories[i]->services[1]->primary            = true;
        accessories[i]->services[1]->type               = HOMEKIT_SERVICE_HEATER_COOLER;
        accessories[i]->services[1]->characteristics    = calloc(8, sizeof(homekit_characteristic_t*));
        accessories[i]->services[1]->characteristics[0] = active;
        accessories[i]->services[1]->characteristics[1] = current_temp;
        accessories[i]->services[1]->characteristics[2] = display_units;
        accessories[i]->services[1]->characteristics[3] = current_state;
        accessories[i]->services[1]->characteristics[4] = target_state;
        accessories[i]->services[1]->characteristics[5] = target_heating_temp;
        accessories[i]->services[1]->characteristics[6] = target_cooling_temp;
        
        // HomeKit Humidity Sensor Characteristics
        homekit_characteristic_t *current_hum           = NEW_HOMEKIT_CHARACTERISTIC(CURRENT_RELATIVE_HUMIDITY, 0);
        
        accessories[i]->services[2]                     = calloc(1, sizeof(homekit_service_t));
        accessories[i]->services[2]->id                 = 16;
        accessories[i]->services[2]->primary            = false;
        accessories[i]->services[2]->type               = HOMEKIT_SERVICE_HUMIDITY_SENSOR;
        accessories[i]->services[2]->characteristics    = calloc(2, sizeof(homekit_characteristic_t*));
        accessories[i]->services[2]->characteristics[0] = current_hum;

        // Creation of thermostat structure data
        homekit_thermostat_t *homekit_thermostat = malloc(sizeof(homekit_thermostat_t));
        memset(homekit_thermostat, 0, sizeof(*homekit_thermostat));

        INFO("[Room %i] Sensor GPIO %i",
             room_index,
             room[room_index].sensor_gpio);
        INFO("[Room %i] Open MCP GPIO %i",
             room_index,
             room[room_index].open_mcp_gpio);
        INFO("[Room %i] Close MCP GPIO %i",
             room_index,
             room[room_index].close_mcp_gpio);
        INFO("[Room %i] Time open-close %i ms",
             room_index,
             room[room_index].time_open_close_ms);
        
        homekit_thermostat->room                = room_index;
        
        homekit_thermostat->sensor_gpio         = room[room_index].sensor_gpio;
        homekit_thermostat->temp_offset         = room[room_index].temp_offset;
        homekit_thermostat->hum_offset          = room[room_index].hum_offset;
        homekit_thermostat->sensor_errors_count = 0;

        homekit_thermostat->open_mcp_gpio       = room[room_index].open_mcp_gpio;
        homekit_thermostat->close_mcp_gpio      = room[room_index].close_mcp_gpio;
        homekit_thermostat->time_open_close_ms  = room[room_index].time_open_close_ms;
        homekit_thermostat->gate_status         = THERMOSTAT_GATE_CLOSED;
        
        homekit_thermostat->active              = active;
        homekit_thermostat->current_temp        = current_temp;
        homekit_thermostat->display_units       = display_units;
        homekit_thermostat->current_state       = current_state;
        homekit_thermostat->target_state        = target_state;
        homekit_thermostat->target_heating_temp = target_heating_temp;
        homekit_thermostat->target_cooling_temp = target_cooling_temp;
        homekit_thermostat->current_hum         = current_hum;
        
        homekit_thermostat->next                = homekit_thermostats;
        homekit_thermostats                     = homekit_thermostat;
        
        sdk_os_timer_setfn(&homekit_thermostat->update_th_timer, update_th_delayed, homekit_thermostat);

        taskYIELD();    // Lets system to do other tasks
    }
    
    sdk_os_timer_setfn(&set_zone_settings_timer, set_zone_settings, NULL);
    
    INFO("\n--------------");
    
    INFO("[WiFi] Init");
    wifi_connect();
    
    INFO("\n[Sensors] Starting...");
    sdk_os_timer_setfn(&read_sensors_timer,
                       read_sensors,
                       NULL);
    sdk_os_timer_arm(&read_sensors_timer,
                     TH_SENSOR_POLL_PERIOD_MS,
                     true);
    read_sensors();
    
    INFO("[WiFi] Waiting for connection...");
    while (sdk_wifi_station_get_connect_status() != STATION_GOT_IP) {
        vTaskDelay(MS_TO_TICK(333));
    }
    
    INFO("[WiFi] Connected");
    
    wifi_channel = sdk_wifi_get_channel();
    INFO("[WiFi] Channel: %i",
         wifi_channel);
    
    INFO("[WiFi] Starting Watchdog...");
    sdk_os_timer_setfn(&wifi_watchdog_timer,
                       wifi_watchdog,
                       NULL);
    sdk_os_timer_arm(&wifi_watchdog_timer,
                     WIFI_WATCHDOG_POLL_PERIOD_MS,
                     true);
    
    INFO("[HomeKit] Starting Server...");
    config.accessories      = accessories;
    config.setupId          = "JOSE";
    config.category         = homekit_accessory_category_other;
    config.config_number    = FIRMWARE_VERSION_OCTAL;
    
#ifdef IAIRZONING_DEBUG
    config.log_output       = true;
#else
    config.log_output       = false;
#endif // IAIRZONING_DEBUG
    
    config.mdns_ttl         = 2500;
    
    homekit_server_init(&config);
    
    INFO("Initial task finished");
    
    vTaskDelete(NULL);
}

/*
 Función inicial del código de usuario. Aquí se crea la tarea de inicio para
 que su ejecución no entorpezca el WatchDog del SDK y no reinicie el sistema por
 una apropiación de la CPU por demasiado tiempo.
 */
void user_init() {
    // Init UART0 for logs output
    uart_set_baud(0, 115200);
    INFO("\n\n\n* iAirZoning *");
    INFO("(c) 2020 José Antonio Jiménez Campos");
    INFO("Universitat Oberta de Catalunya");
    INFO("Grade Final Project\n");
    
    // Disable WiFi
    sdk_wifi_station_set_auto_connect(false);
    sdk_wifi_set_opmode(STATION_MODE);
    sdk_wifi_station_disconnect();

    INFO("Running initial task...");
    xTaskCreate(init_task,
                "init_task",
                INIT_TASK_SIZE,
                NULL,
                INIT_TASK_PRIORITY,
                NULL);
}
