/*
 * iAirZoning
 * Copyright 2020 José Antonio Jiménez Campos
 */

#include <stdio.h>
#include <string.h>
#include <esp8266.h>
#include <espressif/esp_common.h>

#include <esplibs/libmain.h>

#include <i2c/i2c.h>

#include "mcp_23017.h"

#define MCP_ADDR        32

uint8_t mcp_i2c_bus;
uint8_t mcp_a_outs, mcp_b_outs;

/*
 Función de inicialización de la interfaz MCP 23017 usando el bus I2C
 @param i2c_bus         Identificador del bus I2C a usar (0, 1, 2...)
 @param scl_pin         Pin SCL usado para el bus I2C
 @param sda_pin         Pin SDA usado para el bus I2C
 @param init_a_outs     Estado inicial de los GPIO del canal A de la interfaz MSP23017
 @param init_b_outs     Estado inicial de los GPIO del canal B de la interfaz MSP23017
 @return                Número de error. Devuelve 0 si OK.
*/
int mcp_init(const uint8_t i2c_bus,
             const uint8_t scl_pin,
             const uint8_t sda_pin,
             const uint8_t init_a_outs,
             const uint8_t init_b_outs) {
    
    int ret = i2c_init(i2c_bus,
                       scl_pin,
                       sda_pin,
                       I2C_FREQ_100K);
    
    if (ret != 0) {
        return ret;
    }
    
    mcp_i2c_bus = i2c_bus;
    
    i2c_start(mcp_i2c_bus);
    
    mcp_a_outs = init_a_outs;
    uint8_t channel = MCP_CHANNEL_A;
    
    i2c_slave_write(mcp_i2c_bus,
                    MCP_ADDR,
                    &channel,
                    &init_a_outs,
                    1);
    
    mcp_b_outs = init_b_outs;
    channel = MCP_CHANNEL_B;
    
    i2c_slave_write(mcp_i2c_bus,
                    MCP_ADDR,
                    &channel,
                    &init_b_outs,
                    1);
    
    return 0;
}

/*
 Función para actuar sobre los GPIO correspondientes de la interfaz MCP 23017
 @param channel         Canal del GPIO sobre el que se va a actuar (0: canal A, 1: canal B)
 @param gpio            GPIO sobre el que se va a actuar
 @param value           Valor a establecer en el GPIO (0: LOW, 1: HIGH)
 @return                Número de error. Devuelve 0 si OK.
*/
int mcp_gpio_write(const uint8_t channel,
                   const uint8_t gpio,
                   const bool value) {
    uint8_t mcp_outs;
    
    const uint8_t bit = 1 << gpio;
    
    if (channel == MCP_CHANNEL_A) {
        if (value) {
            mcp_a_outs |= bit;
        } else if ((mcp_a_outs & bit) != 0) {
            mcp_a_outs ^= bit;
        }
        
        mcp_outs = mcp_a_outs;
        
    } else if (channel == MCP_CHANNEL_B) {
        if (value) {
            mcp_b_outs |= bit;
        } else if ((mcp_b_outs & bit) != 0) {
            mcp_b_outs ^= bit;
        }
        
        mcp_outs = mcp_b_outs;
        
    } else {
        return -1;
    }
    
    i2c_slave_write(mcp_i2c_bus,
                    MCP_ADDR,
                    &channel,
                    &mcp_outs,
                    1);
    
    return 0;
}
