/*
 * iAirZoning
 * Copyright 2020 José Antonio Jiménez Campos
 */

#ifndef __IAIRZONING_USER_CONFIG_H__
#define __IAIRZONING_USER_CONFIG_H__

/*
 Estructura de datos para introducir la información de los sensores y las compuertas
 sensor_gpio            GPIO del ESP8266 al que va conectado el pin de datos del sensor de temperatura/humedad
 open_mcp_gpio          GPIO de la interfaz MCP23017 al que va conectado el cable de -12V de la compuerta
 close_mcp_gpio         GPIO de la interfaz MCP23017 al que va conectado el cable de +12V de la compuerta
 time_open_close_ms     Tiempo en milisegundos que se necesita para abriri o cerrar la compuerta
 temp_offset            Corrección de la lectura de temperatura del sensor
 hum_offset             Corrección de la lectura de la humedad del sensor
 */
typedef struct room_info {
    uint8_t sensor_gpio;

    uint8_t open_mcp_gpio;
    uint8_t close_mcp_gpio;
    uint16_t time_open_close_ms;
    
    float temp_offset;
    float hum_offset;
} room_info_t;

// WiFi Network Connection Config
#define WIFI_USER_CONFIG_SSID               "mi_wifi"
#define WIFI_USER_CONFIG_PASSWD             "mi_pass"          // Set to NULL if no WiFi password is required (Opened WiFi)

// MCP 23017 Parameters
#define MCP_I2C_BUS                         0
#define MCP_SCL_GPIO                        3
#define MCP_SDA_GPIO                        2
#define INIT_MCP_A_OUTS                     255     // Relays array works inverted
#define INIT_MCP_B_OUTS                     0       // Not used, keep them LOW

// Sensor Parameters
#define TH_SENSOR_POLL_PERIOD_MS            30000
#define TH_SENSOR_MAX_ALLOWED_ERRORS        3

// Thermostat
#define THERMOSTAT_DEADBAND                 0.2f
#define THERMOSTAT_DEADBAND_FORCE_IDLE      0.4f
#define THERMOSTAT_DEADBAND_SOFT_ON         0.4f
#define THERMOSTAT_SET_MIN_TEMP             19.0f
#define THERMOSTAT_SET_MAX_TEMP             28.0f

// Rooms
#define TOTAL_ROOMS                         4

#define ROOM_NINO                           0
#define ROOM_MATRIMONIO                     1
#define ROOM_ESTUDIO                        2
#define ROOM_INVITADOS                      3

#define ROOM_NINO_SENSOR_GPIO               5
#define ROOM_MATRIMONIO_SENSOR_GPIO         4
#define ROOM_ESTUDIO_SENSOR_GPIO            14
#define ROOM_INVITADOS_SENSOR_GPIO          12

#define ROOM_NINO_GATE_OPEN_GPIO            0
#define ROOM_NINO_GATE_CLOSE_GPIO           1
#define ROOM_MATRIMONIO_GATE_OPEN_GPIO      2
#define ROOM_MATRIMONIO_GATE_CLOSE_GPIO     3
#define ROOM_ESTUDIO_GATE_OPEN_GPIO         4
#define ROOM_ESTUDIO_GATE_CLOSE_GPIO        5
#define ROOM_INVITADOS_GATE_OPEN_GPIO       6
#define ROOM_INVITADOS_GATE_CLOSE_GPIO      7

#define GATE_OPEN_CLOSE_TIME_MS             2700
#define SENSOR_TEMPERATURE_OFFSET           0.00f
#define SENSOR_HUMIDITY_OFFSET              -3.00f

const room_info_t const room[TOTAL_ROOMS] = {
    // ROOM_NINO
    { ROOM_NINO_SENSOR_GPIO,
        ROOM_NINO_GATE_OPEN_GPIO,
        ROOM_NINO_GATE_CLOSE_GPIO,
        GATE_OPEN_CLOSE_TIME_MS,
        SENSOR_TEMPERATURE_OFFSET,
        SENSOR_HUMIDITY_OFFSET },

    // ROOM_MATRIMONIO
    { ROOM_MATRIMONIO_SENSOR_GPIO,
        ROOM_MATRIMONIO_GATE_OPEN_GPIO,
        ROOM_MATRIMONIO_GATE_CLOSE_GPIO,
        GATE_OPEN_CLOSE_TIME_MS,
        SENSOR_TEMPERATURE_OFFSET,
        SENSOR_HUMIDITY_OFFSET },
    
    // ROOM_ESTUDIO
    { ROOM_ESTUDIO_SENSOR_GPIO,
        ROOM_ESTUDIO_GATE_OPEN_GPIO,
        ROOM_ESTUDIO_GATE_CLOSE_GPIO,
        GATE_OPEN_CLOSE_TIME_MS,
        SENSOR_TEMPERATURE_OFFSET,
        SENSOR_HUMIDITY_OFFSET },
    
    // ROOM_INVITADOS
    { ROOM_INVITADOS_SENSOR_GPIO,
        ROOM_INVITADOS_GATE_OPEN_GPIO,
        ROOM_INVITADOS_GATE_CLOSE_GPIO,
        GATE_OPEN_CLOSE_TIME_MS,
        SENSOR_TEMPERATURE_OFFSET,
        SENSOR_HUMIDITY_OFFSET }
};

// IrDA Parameters
#define IR_LED_GPIO                         16
#define IR_LED_INVERTED                     0       // 0 = no, 1 = yes
#define IR_LED_FREQ                         38      // KHz

// IrDA Protocol
#define IR_HEADER_MARK                      3400
#define IR_HEADER_SPACE                     1750
#define IR_BIT0_MARK                        450
#define IR_BIT0_SPACE                       420
#define IR_BIT1_MARK                        450
#define IR_BIT1_SPACE                       1300
#define IR_FOOTER_MARK                      440
#define IR_PACKET_SIZE                      275

// IrDA Codes
uint8_t ac_main_ir_off[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 18, 1, 4, 3, 4, 2, 4, 1, 21, 8, 1, 4, 3, 4, 2, 4, 1, 21
};

uint8_t ac_main_ir_heater_on[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 16, 1, 2, 1, 4, 2, 2, 1, 1, 2, 4, 1, 21, 6, 1, 2, 1, 4, 2, 2, 1, 1, 2, 4, 1, 21
};

uint8_t ac_main_ir_heater_idle[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 16, 1, 2, 1, 4, 2, 4, 2, 4, 1, 21, 6, 1, 2, 1, 4, 2, 4, 2, 4, 1, 21
};

uint8_t ac_main_ir_heater_force_idle[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 16, 1, 2, 1, 2, 1, 7, 2, 4, 1, 21, 6, 1, 2, 1, 2, 1, 7, 2, 4, 1, 21
};

uint8_t ac_main_ir_cooler_on[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 16, 1, 1, 1, 3, 2, 4, 1, 1, 2, 4, 1, 21, 6, 1, 1, 1, 3, 2, 4, 1, 1, 2, 4, 1, 21
};

uint8_t ac_main_ir_cooler_idle[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 16, 1, 1, 1, 3, 2, 6, 2, 4, 1, 21, 6, 1, 1, 1, 3, 2, 6, 2, 4, 1, 21
};

uint8_t ac_main_ir_cooler_force_idle[] = {
    0, 2, 3, 1, 2, 2, 1, 1, 2, 2, 1, 2, 2, 1, 2, 1, 4, 1, 16, 1, 1, 1, 4, 3, 4, 2, 4, 1, 21, 6, 1, 1, 1, 4, 3, 4, 2, 4, 1, 21
};

#endif // __IAIRZONING_USER_CONFIG_H__
